import copy
import curses
import os
import random
import time

from .baseUI import TextUI

P1_MARKER = 'X'
P2_MARKER = 'O'
X_STEP = 4
Y_STEP = 2
X_OFFSET = 1
Y_OFFSET = 4

TIE = "                        _\n| |                     | |\n| |                     | |\n| |                     | " \
      "|\n| |                     | |\n| |        .---.        | |\n| |     _.'\   /'._     | |\n| b__--- | .'''. | " \
      "---__d |\n| p__---_| '._.' |_---''q |\n| |       ./   \.       | |\n| |        '---'        | |\n| |           " \
      "          | |\n| |                     | |\n|_|                     |_| "

NOTES_FREQUENCIES = {
    "C4": 261.63,
    "D4": 293.66,
    "D4#": 311.13,
    "E4": 329.63,
    "F4": 349.23,
    "G4": 392,
    "A4": 440,
    "B4": 493.88,
    "C5": 523.25
}


class Game(TextUI):
    """
       This is a class for a 2 player tic-tac-toe game.
       It inherits the TextUI class that is an abstraction for the python curses library

       Attributes:
           game_is_on (bool): Flag that indicates wether a game is been played. Turns off after a victory or a draw.
           is_player1_turn (bool): Flag that controls players turn. First to play is randomly generated.
           total_moves (int): Keeps the number of moves played. Used to check for game ties.
           board_state (list): 3 * 3 array that keeps the game board current layout.
           position_x (int): Keeps the cursor x position on the game board.
           position_y (int): Keeps the cursor y position on the game board.
           window_lines (int): Keeps the terminal total lines of display. Used for centering the game in the screen.
           window_cols (int): Keeps the terminal total columns of display. Used for centering the game in the screen.
           plays (array): Keeps all the board_states in a game.
       """

    def __init__(self):
        super().__init__()
        self.game_is_on = True
        self.is_player1_turn = bool(random.getrandbits(1))
        self.total_moves = 0
        self.board_state = [list('   ') for _ in range(3)]
        self.position_x = 1
        self.position_y = 1
        self.window_lines = 0
        self.window_cols = 0
        self.plays = []
        self.title = "### TIC-TAC-TOE ###"
        self.instructions = "Use 'arrow keys' to move,  'Space' to place marker,  'Q' to Quit"

    def input(self, key):
        """
            Handles player inputs.

            Runs the game depending on pressed key.

            :param key: String representation of pressed key.

            :return: nothing
        """

        # Quit the game anytime
        if key in ["q", "Q"] or key and not self.game_is_on:
            self.stop()

        # Get coordinates to place elements based on window size
        middle_y = int(self.window_lines / 2)
        middle_x = int(self.window_cols / 2)

        # Move cursor or place marker accordingly to input
        if key == "KEY_UP":
            self.position_y = max(0, self.position_y - 1)
        elif key == "KEY_DOWN":
            self.position_y = min(2, self.position_y + 1)
        elif key == "KEY_LEFT":
            self.position_x = max(0, self.position_x - 1)
        elif key == "KEY_RIGHT":
            self.position_x = min(2, self.position_x + 1)
        elif key == " ":
            # Get cursor position
            y, x = self.screen.getyx()

            # Don't place marker if position already has a marker. Play a warning sound instead.
            if self.screen.inch(y, x) == ord(' '):
                # Place a marker, play player sound, change player turn, update game total moves and save board state.
                self.place_marker()

                player1_sound() if self.is_player1_turn else player2_sound()
                self.is_player1_turn = not self.is_player1_turn
                self.total_moves += 1
                self.plays.append(copy.deepcopy(self.board_state))

                # Check if someone has won or if it's a tie |-o-|
                if self.someone_has_won():

                    winning_message = f"### Player {P2_MARKER if self.is_player1_turn else P1_MARKER} wins ###"
                    self.screen.clear()
                    self.screen.addstr(middle_y, middle_x - int(len(winning_message) / 2),
                                       winning_message,
                                       curses.A_BLINK)
                    victory_sound()
                    self.end_game()

                elif self.total_moves > 8:
                    self.screen.clear()
                    tie_image = f"### It's a tie. ###\n {TIE}"
                    self.screen.addstr(middle_y, middle_x, tie_image, curses.A_BLINK)
                    tie_sound()
                    self.end_game()

            else:
                curses.beep()

    def end_game(self):

        """
        Ends current game.

        Called when a player wins or when a tie is reached.
        It's called continuously. Check TextUI class main loop.


        :return: Nothing
        """

        self.game_is_on = False
        # One last draw to render the final play and the game end message.
        self.draw()
        # Plays have ended so we don0t need the cursor blinking on screen
        curses.curs_set(0)
        # Let the player view the final game state for as long as he/she wants.
        self.pause()

    def draw(self):

        """
        Handles the rendering of all game components into the screen.

        It's called continuously. Check TextUI class main loop.

        :return: Nothing
        """

        # self.screen.addstr(20, 20, str(self.board_state))
        # self.screen.addstr(22, 20, str(self.total_moves))
        # self.screen.addstr(24, 20, str(self.game_is_on))
        # self.screen.addstr(26, 20, str(self.plays))

        # Nice border around the terminal
        self.screen.border()

        # How big is the terminal screen?
        self.window_lines = curses.LINES
        self.window_cols = curses.COLS

        # Where we will draw the board and title
        top_y = int(self.window_lines / 4)
        middle_x = int(self.window_cols / 2)

        # Adding title and instructions
        self.screen.addstr(1, middle_x - int(len(self.title) / 2), self.title)
        self.screen.addstr(3, middle_x - int(len(self.instructions) / 2), self.instructions)

        # Add the board.
        self.screen.addstr(top_y, middle_x,
                           f'{self.board_state[0][0]} │ {self.board_state[0][1]} │ {self.board_state[0][2]} ')
        self.screen.addstr(top_y + 1, middle_x, '──┼───┼──')
        self.screen.addstr(top_y + 2, middle_x,
                           f'{self.board_state[1][0]} │ {self.board_state[1][1]} │ {self.board_state[1][2]} ')
        self.screen.addstr(top_y + 3, middle_x, '──┼───┼──')
        self.screen.addstr(top_y + 4, middle_x,
                           f'{self.board_state[2][0]} │ {self.board_state[2][1]} │ {self.board_state[2][2]} ')

        # Player info including whose turn is it.
        self.show_players_info()

        # Move cursor to new position
        self.screen.move(top_y + self.position_y * Y_STEP, middle_x + self.position_x * X_STEP)

    def show_players_info(self):
        """
        Show whose turn it is, only when the game is on.

        :return: Nothing
        """
        if self.game_is_on:
            self.screen.addstr(Y_OFFSET + 6, 1, f'Player {P1_MARKER}',
                               curses.A_BLINK if self.is_player1_turn else 0)
            self.screen.addstr(Y_OFFSET + 7, 1, f'Player {P2_MARKER}',
                               curses.A_BLINK if not self.is_player1_turn else 0)

    def place_marker(self):

        """
        Marks the board with current player mark.

        :return: Nothing
        """

        self.board_state[self.position_y][self.position_x] = P1_MARKER if self.is_player1_turn else P2_MARKER

    def someone_has_won(self):
        """
        Checks, after each play, if someone has won.

        :return: True if a winning pattern is present on the board. False otherwise.
        """

        # Check last move for horizontal victory
        if self.board_state[0][self.position_x] \
                == self.board_state[1][self.position_x] \
                == self.board_state[2][self.position_x]:
            return True
        # Check last move for vertical victory
        if self.board_state[self.position_y][0] \
                == self.board_state[self.position_y][1] \
                == self.board_state[self.position_y][2]:
            return True

        # Check last move for main diagonal victory i.e. where position_y == position_x
        if self.position_y == self.position_x and self.board_state[0][0] \
                == self.board_state[1][1] \
                == self.board_state[2][2]:
            return True
        # Check last move for secondary diagonal victory i.e. where position_x + position_y == 2
        if self.position_y + self.position_x == 2 and self.board_state[0][2] \
                == self.board_state[1][1] \
                == self.board_state[2][0]:
            return True

        return False


def beep(note, duration):
    """
    Plays a sound of the given 'note' with a given 'duration'

    :param note: (string) A representation of a musical note. Check NOTES_FREQUENCIES for available notes.
    :param duration: (int) The duration expressed in milliseconds.

    :return: Nothing
    """
    frequency = get_frequency(note)
    os.system('play --no-show-progress -n -t alsa synth %s sin %s' % (duration / 1000, frequency))


def get_frequency(note):
    """
    Returns the frequency in Hz of the given note. Check NOTES_FREQUENCIES for available notes.

    :param note: (string) A representation of a musical note. Check NOTES_FREQUENCIES for available notes.

    :return: (float) Frequency of the given 'note'
    """
    frequency = NOTES_FREQUENCIES[note]
    return frequency


def player1_sound():
    """
    Player 1 marking sound
    :return: Nothing
    """
    beep("G4", 100)
    beep("C5", 50)


def player2_sound():
    """
    Player 2 marking sound
    :return: Nothing
    """
    beep("C5", 50)
    beep("G4", 100)


def play_intro():
    """
    Game intro sound

    :return: Nothing
    """
    beep("G4", 120)
    beep("G4", 120)
    beep("G4", 120)
    beep("D4#", 600)
    time.sleep(0.3)
    beep("F4", 120)
    beep("F4", 120)
    beep("F4", 120)
    beep("D4", 600)


def victory_sound():
    """
      Victory sound

      :return: Nothing
    """
    beep("G4", 80)
    beep("G4", 80)
    beep("G4", 80)
    beep("C5", 160)
    beep("G4", 80)
    beep("C5", 300)


def tie_sound():
    """
      Tie sound

      :return: Nothing
    """
    time.sleep(0.3)
    beep("C5", 160)
    time.sleep(0.1)
    beep("C4", 80)
    beep("C4", 80)
